diff --git a/MyComponent.tsx b/MyComponent.tsx
new file mode 100644
index 0000000..f4abee1
--- /dev/null
+++ b/MyComponent.tsx
@@ -0,0 +1,8 @@
+export function MyComponent() {
+  return (
+    <section>
+      <h2>About me</h2>
+      <p>I'm a custom JSX component.</p>
+    </section>
+  );
+}
diff --git a/deno.json b/deno.json
new file mode 100644
index 0000000..a759202
--- /dev/null
+++ b/deno.json
@@ -0,0 +1,12 @@
+{
+  "tasks": {
+    "build": "deno run --allow-write main.ts"
+  },
+  "compilerOptions": {
+    "jsx": "react-jsx",
+    "jsxImportSource": "react"
+  },
+  "imports": {
+    "react/jsx-runtime": "./react.ts"
+  }
+}
diff --git a/index.tsx b/index.tsx
new file mode 100644
index 0000000..0f9b5a1
--- /dev/null
+++ b/index.tsx
@@ -0,0 +1,15 @@
+import { MyComponent } from "./MyComponent.tsx";
+
+export default function () {
+  return (
+    <>
+      <section>
+        <h1>Output</h1>
+        <p>
+          This page was dynamically generated on {new Date().toUTCString()}.
+        </p>
+      </section>
+      <MyComponent />
+    </>
+  );
+}
diff --git a/main.ts b/main.ts
new file mode 100644
index 0000000..e8ec8da
--- /dev/null
+++ b/main.ts
@@ -0,0 +1,3 @@
+import page from "./index.tsx";
+
+Deno.writeTextFileSync("index.html", String(page()));
diff --git a/react.d.ts b/react.d.ts
new file mode 100644
index 0000000..71fb3bb
--- /dev/null
+++ b/react.d.ts
@@ -0,0 +1,8 @@
+declare namespace JSX {
+  interface Element {
+    [attribute: string]: string | number | boolean;
+  }
+  interface IntrinsicElements {
+    [elemName: string]: Element;
+  }
+}
diff --git a/react.ts b/react.ts
new file mode 100644
index 0000000..8aee7ca
--- /dev/null
+++ b/react.ts
@@ -0,0 +1,71 @@
+const VOID_TAGS = [
+  "area",
+  "base",
+  "br",
+  "col",
+  "embed",
+  "hr",
+  "img",
+  "input",
+  "link",
+  "meta",
+  "param",
+  "source",
+  "track",
+  "wbr",
+];
+
+type JSXProps = {
+  children: unknown | unknown[];
+  [k: string]: unknown;
+};
+
+export function jsxs(
+  tag: unknown,
+  props: JSXProps,
+  _key?: unknown,
+): string {
+  if (typeof tag === "function") { // custom component
+    return tag(props);
+  } else if (typeof tag === "string") { // native HTML element
+    const attrs = Object.entries(props).reduce<string[]>(
+      (acc, [key, value]) => {
+        if (key === "children") {
+          // ignore the children attribute
+        } else if (typeof value === "boolean") {
+          // keep true attrs unchanged and remove false attrs
+          if (value) {
+            acc.push(`${key}`);
+          }
+        } else {
+          acc.push(`${key}="${value}"`);
+        }
+        return acc;
+      },
+      [""],
+    ).join(" ");
+
+    return VOID_TAGS.includes(tag)
+      ? `<${tag}${attrs}>`
+      : `<${tag}${attrs}>${appendChild(props.children)}</${tag}>`;
+  } else {
+    return "";
+  }
+}
+
+export const jsx = jsxs;
+
+export function Fragment(props: JSXProps): string {
+  return appendChild(props.children);
+}
+
+function appendChild(child: unknown): string {
+  if (child === false || child === null || child === undefined) {
+    return "";
+  }
+
+  if (Array.isArray(child)) {
+    return child.map(appendChild).join("");
+  }
+  return String(child);
+}
